package me.hammerle.code;

import java.lang.reflect.InvocationTargetException;
import me.hammerle.exceptions.HoldCodeException;
import java.math.BigDecimal;
import java.math.RoundingMode;
import java.time.ZonedDateTime;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Calendar;
import java.util.Collection;
import java.util.Collections;
import java.util.GregorianCalendar;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.function.BiFunction;
import java.util.function.BiConsumer;
import java.util.function.Consumer;
import java.util.stream.Collectors;
import me.hammerle.exceptions.PreScriptException;
import me.hammerle.exceptions.SnuviException;

public class SnuviParser 
{
    protected final ISnuviLogger logger;
    private final ISnuviScheduler scheduler;
    private final HashMap<String, BiFunction<Object[], Script, Object>> functions;
    public boolean printStack;
    
    private int idCounter;
    private final HashMap<Integer, Script> scripts;
    private final ArrayList<Integer> termQueue;
    
    public SnuviParser(ISnuviLogger logger, ISnuviScheduler scheduler)
    {
        this.logger = logger;
        this.scheduler = scheduler;
        functions = new HashMap<>(); 
        registerStandardLibraries();
        
        printStack = false;
        
        scripts = new HashMap<>();
        termQueue = new ArrayList<>();
        idCounter = 0;
    }
    
    // -----------------------------------------------------------------------------------
    // function registry
    // -----------------------------------------------------------------------------------
    
    public boolean isRegisteredFunction(String s)
    {
        return functions.containsKey(s);
    }
    
    public void registerFunction(String s, BiFunction<Object[], Script, Object> f)
    {
        functions.put(s, f);
    }
    
    public void registerConsumer(String s, BiConsumer<Object[], Script> f)
    {
        functions.put(s, (BiFunction<Object[], Script, Object>) (args, sc) -> 
        {
            f.accept(args, sc);
            return Void.TYPE;
        });
    }
    
    public void registerAlias(String alias, String original)
    {
        BiFunction<Object[], Script, Object> f = functions.get(original);
        if(f == null)
        {
            throw new IllegalArgumentException("registering alias for non existent function");
        }
        registerFunction(alias, functions.get(original)); 
    }
    
    @SuppressWarnings("unchecked")
    private void registerStandardLibraries()
    {
        registerFunction("nothing", (args, sc) -> 
                0);
        registerFunction("", (args, sc) -> 
                args[0]);
        registerConsumer("error", (args, sc) -> 
                printStack = !printStack);       
        
        // -------------------------------------------------------------------------------    
        // event
        // -------------------------------------------------------------------------------  
        registerConsumer("event.load", (args, sc) -> 
                sc.loadEvent(args[0].toString())); 
        registerConsumer("event.unload", (args, sc) -> 
                sc.unloadEvent(args[0].toString())); 
        registerFunction("event.isloaded", (args, sc) -> 
                sc.isLoadedEvent(args[0].toString())); 

        // -------------------------------------------------------------------------------    
        // bit
        // ------------------------------------------------------------------------------- 
        
        registerFunction("bit.rightshift", (args, qd) -> 
                (double) (getInt(args[0]) >> getInt(args[1])));
        registerFunction("bit.leftshift", (args, qd) -> 
                (double) (getInt(args[0]) << getInt(args[1])));
        registerFunction("bit.and", (args, qd) -> 
                (double) (getInt(args[0]) & getInt(args[1])));
        registerFunction("bit.or", (args, qd) -> 
                (double) (getInt(args[0]) | getInt(args[1])));
        registerFunction("bit.xor", (args, qd) -> 
                (double) (getInt(args[0]) ^ getInt(args[1])));
        registerFunction("bit.invert", (args, qd) -> 
                (double) (~getInt(args[0])));
        registerFunction("bit.set", (args, qd) -> 
                (double) (getInt(args[0]) | (getInt(args[2]) << getInt(args[1]))));
        registerFunction("bit.get", (args, qd) -> 
                (double) ((getInt(args[0]) & (1 << getInt(args[1])))) != 0);
        
        // -------------------------------------------------------------------------------    
        // math
        // -------------------------------------------------------------------------------    
        registerFunction("math.mod", (args, sc) -> 
                ((double) args[0]) % ((double) args[1]));
        registerFunction("math.abs", (args, sc) -> 
                Math.abs((double) args[0]));
        registerFunction("math.pow", (args, sc) -> 
                Math.pow((double) args[0], (double) args[1]));
        registerFunction("math.root", (args, sc) -> 
                Math.pow((double) args[0], 1d / ((double) args[1])));
        registerFunction("math.sin", (args, sc) -> 
                Math.sin((double) args[0]));
        registerFunction("math.cos", (args, sc) -> 
                Math.cos((double) args[0]));
        registerFunction("math.tan", (args, sc) -> 
                Math.tan((double) args[0]));
        registerFunction("math.asin", (args, sc) -> 
                Math.asin((double) args[0]));
        registerFunction("math.acos", (args, sc) -> 
                Math.acos((double) args[0]));
        registerFunction("math.atan", (args, sc) -> 
                Math.atan((double) args[0]));
        registerFunction("math.e", (args, sc) -> 
                Math.E);
        registerFunction("math.pi", (args, sc) -> 
                Math.PI);
        registerFunction("math.ln", (args, sc) -> 
                Math.log((double) args[0]));
        registerFunction("math.log", (args, sc) -> 
                Math.log10((double) args[0]));
        registerFunction("math.random", (args, sc) -> 
                (double) ScriptUtils.randomInt(getInt(args[0]), getInt(args[1])));
        registerFunction("math.round", (args, sc) -> 
                Math.round((double) args[0]));
        registerFunction("math.rounddown", (args, sc) -> 
                Math.floor((double) args[0]));
        registerFunction("math.roundup", (args, sc) -> 
                Math.ceil((double) args[0]));
        registerFunction("math.roundcomma", (args, sc) -> 
                new BigDecimal(((double) args[0])).setScale(getInt(args[1]), RoundingMode.HALF_UP).doubleValue());
        
        // -------------------------------------------------------------------------------  
        // lists
        // -------------------------------------------------------------------------------    
        registerConsumer("list.new", (args, sc) ->                                                
                sc.setVar(args[0].toString(), new ArrayList<>()));
        registerFunction("list.exists", (args, sc) ->                                                
                args[0] instanceof List); 
        registerConsumer("list.add", (args, sc) ->                            
                ((List) args[0]).add(args[1]));
        registerConsumer("list.remove", (args, sc) ->                            
                ((List) args[0]).remove(args[1]));
        registerConsumer("list.removeindex", (args, sc) ->                            
                ((List) args[0]).remove(getInt(args[1])));                
        registerFunction("list.contains", (args, sc) ->                            
                ((List) args[0]).contains(args[1]));
        registerFunction("list.getsize", (args, sc) ->                            
                (double) ((List) args[0]).size());
        registerFunction("list.getindex", (args, sc) ->                            
                ((List) args[0]).get(getInt(args[1])));
        registerConsumer("list.setindex", (args, sc) ->                            
                ((List) args[0]).set(getInt(args[1]), args[2]));
        registerConsumer("list.clear", (args, sc) ->                            
                ((List) args[0]).clear());
        registerFunction("list.getindexof", (args, sc) ->                            
                (double) ((List) args[0]).indexOf(args[1]));  
        registerConsumer("list.sort", (args, sc) ->  
                sortList((List<Object>) args[0], sc));
        registerConsumer("list.reverse", (args, sc) ->                            
                Collections.reverse((List<Object>) args[0])); 
        registerConsumer("list.shuffle", (args, sc) ->                            
                Collections.shuffle((List<Object>) args[0]));

        // -------------------------------------------------------------------------------  
        // maps
        // ------------------------------------------------------------------------------- 
        registerConsumer("map.new", (args, sc) ->                                                
                sc.setVar(args[0].toString(), new HashMap<>())); 
        registerFunction("map.exists", (args, sc) ->                                                
                args[0] instanceof Map);  
        registerConsumer("map.add", (args, sc) ->                            
                ((HashMap) args[0]).put(args[1], args[2]));
        registerConsumer("map.remove", (args, sc) ->                            
                ((HashMap) args[0]).remove(args[1]));
        registerFunction("map.contains", (args, sc) ->                            
                ((HashMap) args[0]).containsKey(args[1]));
        registerFunction("map.getsize", (args, sc) ->                            
                (double) ((HashMap) args[0]).size());
        registerFunction("map.get", (args, sc) ->                            
                ((HashMap) args[0]).get(args[1]));
        registerConsumer("map.clear", (args, sc) ->                            
                ((HashMap) args[0]).clear());
        
        // -------------------------------------------------------------------------------  
        // sets
        // ------------------------------------------------------------------------------- 
        registerConsumer("set.new", (args, sc) ->                                                
                sc.setVar(args[0].toString(), new HashSet<>())); 
        registerFunction("set.exists", (args, sc) ->                                                
                args[0] instanceof Set);  
        registerConsumer("set.add", (args, sc) ->                            
                ((HashSet) args[0]).add(args[1]));
        registerConsumer("set.remove", (args, sc) ->                            
                ((HashSet) args[0]).remove(args[1]));
        registerFunction("set.contains", (args, sc) ->                            
                ((HashSet) args[0]).contains(args[1]));
        registerFunction("set.getsize", (args, sc) ->                            
                (double) ((HashSet) args[0]).size());

        // -------------------------------------------------------------------------------  
        // time
        // -------------------------------------------------------------------------------
        registerFunction("time.get", (args, sc) ->                            
                getDouble(System.currentTimeMillis()));
        registerFunction("time.nextday", (args, sc) ->         
                getDouble(getNextDay(args)));   
        registerFunction("time.getyear", (args, sc) ->         
                (double) getYear(args));   
        registerFunction("time.getmonth", (args, sc) ->         
                (double) getMonth(args));   
        registerFunction("time.getday", (args, sc) ->         
                (double) getDay(args));   
        registerFunction("time.gethour", (args, sc) ->         
                (double) getHour(args));   
        registerFunction("time.getminute", (args, sc) ->         
                (double) getMinute(args));   
        registerFunction("time.getsecond", (args, sc) ->         
                (double) getSecond(args));                   
              
        // -------------------------------------------------------------------------------    
        // text
        // -------------------------------------------------------------------------------   
        registerFunction("text.number", (args, sc) ->         
                doubleToString(getDouble(args[0]))); 
        registerFunction("text.class", (args, sc) ->         
                args[0].getClass().getSimpleName()); 
        
        // -------------------------------------------------------------------------------    
        // commands without library
        // -------------------------------------------------------------------------------    
        registerFunction("add", (args, sc) -> 
                Arrays.stream(args).mapToDouble(s -> (double) s).sum());
        registerFunction("sub", (args, sc) -> 
                ((double) args[0]) - ((double) args[1]));
        registerFunction("inc", (args, sc) -> 
                increaseVar(args[0], sc, 1)); 
        registerFunction("dec", (args, sc) -> 
                increaseVar(args[0], sc, -1));
        registerFunction("mul", (args, sc) -> 
                ((double) args[0]) * ((double) args[1]));
        registerFunction("div", (args, sc) -> 
                ((double) args[0]) / ((double) args[1]));
        registerFunction("getvar", (args, sc) -> 
                sc.getVar(args[0].toString()));
        registerConsumer("setvar", (args, sc) -> 
                sc.setVar(args[0].toString(), args[1]));
        registerConsumer("removevar", (args, sc) -> 
                sc.removeVar(args[0].toString()));
        registerConsumer("reset", (args, sc) -> 
                sc.resetLoopCounter());
        registerConsumer("wait", (args, sc) -> 
                { sc.resetLoopCounter(); throw new HoldCodeException(); });
        registerConsumer("goto", (args, sc) -> 
                sc.gotoLabel(args[0].toString(), false));
        registerConsumer("gotoline", (args, sc) -> 
                { sc.gotoSpecialJumpLine(); sc.incLoopCounter(); });
        registerConsumer("sgoto", (args, sc) -> 
                scheduleGoto(args, sc));
        registerConsumer("gosub", (args, sc) -> 
                sc.gotoLabelWithReturn(args[0].toString()));
        registerConsumer("return", (args, sc) -> 
                sc.doReturn());
        registerConsumer("try", (args, sc) -> 
                sc.saveTryJumpLine());                   
        registerConsumer("catch", (args, sc) -> 
                sc.gotoSpecialJumpLine());
        registerConsumer("if", (args, sc) -> 
                ifFunction(args, sc));   
        registerConsumer("else", (args, sc) -> 
                sc.gotoSpecialJumpLine());   
        registerConsumer("while", (args, sc) -> 
                whileFunction(args, sc)); 
        registerFunction("equal", (args, sc) -> 
                isEqual(args));
        registerAlias("equals", "equal");
        registerFunction("less", (args, sc) -> 
                (double) args[0] < (double) args[1]);
        registerFunction("greater", (args, sc) -> 
                (double) args[0] > (double) args[1]);
        registerFunction("notequal", (args, sc) -> 
                !isEqual(args));
        registerFunction("lessequal", (args, sc) -> 
                (double) args[0] <= (double) args[1]);
        registerFunction("greaterequal", (args, sc) -> 
                (double) args[0] >= (double) args[1]);
        registerFunction("invert", (args, sc) -> 
                !((boolean) args[0]));
        registerFunction("and", (args, sc) -> 
                Arrays.stream(args).allMatch(s -> s.equals(true)));
        registerFunction("or", (args, sc) -> 
                Arrays.stream(args).anyMatch(s -> s.equals(true)));
        registerFunction("concatlist", (args, sc) ->      
                ((List<Object>) args[0]).stream().limit(getInt(args[3]) + 1).skip(getInt(args[2])).map(o -> o.toString()).collect(Collectors.joining(args[1].toString())));
        registerConsumer("split", (args, sc) ->      
                split(args, sc));             
        registerFunction("concat", (args, sc) ->                            
                ScriptUtils.connect(args, 0));
        registerFunction("tolowercase", (args, sc) ->                            
                ScriptUtils.connect(args, 0).toLowerCase());
        registerFunction("touppercase", (args, sc) ->                            
                ScriptUtils.connect(args, 0).toUpperCase());   
        registerConsumer("waitfor", (args, sc) ->     
                waitFor(args, sc));
        registerConsumer("term", (args, sc) -> 
                { termSafe(sc); throw new HoldCodeException(); }); 
    }
    
    @SuppressWarnings("unchecked")
    protected Object parseFunction(Script sc, String function, Object[] args) throws HoldCodeException
    {
        try
        {
            return functions.get(function).apply(args, sc);
        }
        catch(HoldCodeException ex)
        {
            throw ex;
        }
        catch(Exception ex)
        {
            if(printStack)
            {
                ex.printStackTrace();
            }
            if(sc.getTryJumpLine() != -1)
            {
                sc.setVar("error", ex.getClass().getSimpleName());
                sc.gotoTryJumpLine();
                sc.resetTryJumpLine();
                return Void.TYPE;
            }
            if(ex instanceof SnuviException)
            {
                logger.printException((SnuviException) ex);
            }
            else
            {
                logger.printException(new SnuviException(ex, sc));
            }
            sc.resetLoopCounter();
            throw new HoldCodeException();
        }
    }
    
    // -----------------------------------------------------------------------------------
    // script controller
    // -----------------------------------------------------------------------------------
    
    public Script getScript(int id)
    {
        return scripts.get(id);
    }
    
    public boolean termUnsafe(Script sc)
    {
        return scripts.remove(sc.getId()) != null;
    }
    
    public void termSafe(Script sc)
    {
        if(sc == null)
        {
            return;
        }
        sc.setInvalid();
        termQueue.add(sc.getId());
    }
    
    private void term()
    {
        if(!termQueue.isEmpty())
        {
            termQueue.forEach(i -> scripts.remove(i));
            termQueue.clear();
        }
    }
    
    public void termAllUnsafe()
    {
        scripts.values().forEach(sc -> sc.setInvalid());
        scripts.clear();
    }
    
    public Collection<Script> getScripts()
    {
        return scripts.values();
    }
    
    private Script startScript(Script script, Object... o)
    { 
        try
        {            
            script.initExpansion(o);
            scripts.put(idCounter, script);
            idCounter++;
            script.runCode();
            term();
            return script;
        }
        catch(PreScriptException ex)
        {
            logger.printException(ex);
            return null;
        }
    }
    
    public Script startScript(String scriptName, String code)
    { 
        return startScript(new Script(this, idCounter, scriptName, code));
    }
    
    public <T extends Script> T startScript(Class<T> c, String scriptName, String code, boolean b, Object... o)
    { 
        try
        {
            return (T) startScript(c.getDeclaredConstructor(
                    SnuviParser.class, int.class, String.class, String.class, boolean.class)
                    .newInstance(this, idCounter, scriptName, code, b), o);
        }
        catch(IllegalAccessException | IllegalArgumentException | InstantiationException |
                InvocationTargetException | NoSuchMethodException | SecurityException ex)
        {
            if(ex instanceof InvocationTargetException)
            {
                Throwable t = ((InvocationTargetException) ex).getCause();
                if(t != null && t instanceof SnuviException)
                {
                    logger.printException((SnuviException) t);
                }
            }
            return null;
        }
    }
    
    // -----------------------------------------------------------------------------------
    // event
    // -----------------------------------------------------------------------------------
    
    public void callEvent(String name, Consumer<Script> before, Consumer<Script> after)
    {
        scripts.values().stream()
                .filter(script -> script.receiveEventBroadcast)
                .filter(script -> script.isLoadedEvent(name))
                .forEach(sc -> 
                {
                    sc.setVar("event", name);
                    if(before != null)
                    {
                        before.accept(sc);
                    }
                    sc.runCode();
                    if(after != null)
                    {
                        after.accept(sc);
                    }
                    sc.eventVars.forEach(s -> sc.removeVar(s));
                    sc.eventVars.clear();
                });
        term();
    }
    
    public boolean callEvent(String name, Script sc, Consumer<Script> before, Consumer<Script> after)
    {
        if(sc.isLoadedEvent(name))
        {
            sc.setVar("event", name);
            if(before != null)
            {
                before.accept(sc);
            }
            sc.runCode();
            if(after != null)
            {
                after.accept(sc);
            }
            sc.eventVars.forEach(s -> sc.removeVar(s));
            sc.eventVars.clear();
            term();
            return true;
        }
        return false;
    }
    
    // -----------------------------------------------------------------------------------
    // functions
    // -----------------------------------------------------------------------------------
    
    private String doubleToString(Double d)
    {
        if(d.doubleValue() == d.longValue())
        {
            return String.valueOf(d.longValue());
        }
        return d.toString();
    }
    
    private void ifFunction(Object[] args, Script sc)
    {
        if(Arrays.stream(args).anyMatch(s -> !((boolean) s)))
        {
            sc.gotoSpecialJumpLine();
            sc.jumpNextIfElse();
        }
    }  
    
    private void whileFunction(Object[] args, Script sc)
    {
        if(Arrays.stream(args).anyMatch(s -> !((boolean) s)))
        {
            sc.gotoSpecialJumpLine();
        }
    } 
                  
    @SuppressWarnings("unchecked")
    private void sortList(List<Object> args, Script sc) 
    {
        Collections.sort(args, (Object o1, Object o2) -> ((Comparable) o1).compareTo(o2));
    }
    
    private void scheduleGoto(Object[] args, Script sc)
    {
        scheduler.scheduleTask(() -> 
        {
            if(!sc.isValid())
            {
                return;
            }
            try
            {
                sc.gotoLabel(args[1].toString(), true);
                sc.setHalt(false);
                sc.runCode();
            }
            catch(Exception ex)
            {
                logger.printException(new SnuviException(ex, sc.getName(), "scheduled goto"));
            }
        }, getInt(args[0]));
    }
    
    @SuppressWarnings("")
    private void waitFor(Object[] args, Script sc) throws UnsupportedOperationException
    {           
        sc.resetLoopCounter();
        long l = getLong(args[0]);
        if(l < 0)
        {
            throw new IllegalArgumentException("time units can't be negative");
        }
        sc.setHalt(true);
        scheduler.scheduleTask(() -> 
        {                   
            if(sc == null || !sc.isValid())
            {
                return;
            }
            sc.setHalt(false);
            sc.runCode();
        }, l); 
    }   
    
    private Number increaseVar(Object var, Script sc, int value)
    {
        double n = ((double) sc.getVar(var.toString())) + value;
        sc.setVar(var.toString(), n);
        return n;
    }   
    
    private static boolean isEqual(Object[] args)
    {
        if(args[0] == null)
        {
            return args[1] == null;
        }
        else if(args[1] == null)
        {
            return false;
        }
        // this is needed for comparing different number types (e.g. Integer and Double)
        // allthough there should be no numbers except doubles...
        else if(args[1] instanceof Number && args[0] instanceof Number)
        {
            return ((Number) args[0]).doubleValue() == ((Number) args[1]).doubleValue();
        }
        return args[0].equals(args[1]);
    }
    
    private void split(Object[] args, Script sc)
    {
        sc.setVar(args[0].toString(), 
                Arrays.stream(ScriptUtils.connect(args, 2).split(args[1].toString()))
                        .map(s -> Code.convertInput(s, false)).collect(Collectors.toList()));
    }
    
    // -----------------------------------------------------------------------------------   
    // number handlers, cause primitive types transform into their classes all the time
    // -----------------------------------------------------------------------------------
    
    private int getInt(Object o)
    {
        return ((Number) o).intValue();
    }
    
    private long getLong(Object o)
    {
        return ((Number) o).longValue();
    }
    
    private double getDouble(Object o)
    {
        return ((Number) o).doubleValue();
    }
    
    // -----------------------------------------------------------------------------------    
    // time handler
    // -----------------------------------------------------------------------------------
    
    private long getNextDay(Object[] args)       
    {
        GregorianCalendar cal = GregorianCalendar.from(ZonedDateTime.now());
        cal.setTimeInMillis(getLong(args[0]));
        cal.add(Calendar.DAY_OF_YEAR, 1);
        cal.set(Calendar.HOUR, 0);
        cal.set(Calendar.SECOND, 0);
        cal.set(Calendar.MINUTE, 0);
        cal.set(Calendar.MILLISECOND, 0);
        return cal.getTimeInMillis();   
    } 
    
    private int getYear(Object[] args)       
    {
        GregorianCalendar cal = GregorianCalendar.from(ZonedDateTime.now());
        cal.setTimeInMillis(getLong(args[0]));
        return cal.get(Calendar.YEAR);   
    }
    
    private int getMonth(Object[] args)       
    {
        GregorianCalendar cal = GregorianCalendar.from(ZonedDateTime.now());
        cal.setTimeInMillis(getLong(args[0]));
        return cal.get(Calendar.MONTH) + 1;   
    }
    
    private int getDay(Object[] args)       
    {
        GregorianCalendar cal = GregorianCalendar.from(ZonedDateTime.now());
        cal.setTimeInMillis(getLong(args[0]));
        return cal.get(Calendar.DAY_OF_MONTH);   
    }
    
    private int getHour(Object[] args)       
    {
        GregorianCalendar cal = GregorianCalendar.from(ZonedDateTime.now());
        cal.setTimeInMillis(getLong(args[0]));
        return cal.get(Calendar.HOUR_OF_DAY);   
    }
    
    private int getMinute(Object[] args)       
    {
        GregorianCalendar cal = GregorianCalendar.from(ZonedDateTime.now());
        cal.setTimeInMillis(getLong(args[0]));
        return cal.get(Calendar.MINUTE);   
    }
    
    private int getSecond(Object[] args)       
    {
        GregorianCalendar cal = GregorianCalendar.from(ZonedDateTime.now());
        cal.setTimeInMillis(getLong(args[0]));
        return cal.get(Calendar.SECOND);   
    }
}