package me.hammerle.math;

public class Fraction implements Cloneable, Comparable<Fraction>
{
    private long numerator;
    private long denominator;
    
    public Fraction(long n, long d)
    {
        if(d == 0)
        {
            throw new ArithmeticException();
        }
        else if(d < 0)
        {
            this.denominator = -d;
            this.numerator = -n;
        }
        else
        {
            this.denominator = d;
            this.numerator = n;
        }
        shortFraction();
    }
    
    public Fraction(long n)
    {
        this(n, 1);
    }
    
    // -----------------------------------------------------------------------------------
    // Konstanten
    // ----------------------------------------------------------------------------------- 
    
    private static final Fraction PI = new Fraction(355, 113);
    private static final Fraction E = new Fraction(2721, 1001);
    
    public static Fraction getPi()
    {
        return PI.clone();
    }
    
    public static Fraction getE()
    {
        return E.clone();
    }
    
    // -----------------------------------------------------------------------------------
    // Ausgabe
    // ----------------------------------------------------------------------------------- 

    public double getDoubleValue()
    {
        return numerator / (double) denominator;
    }
    
    public long getLongValue()
    {
        return numerator / denominator;
    }
    
    // -----------------------------------------------------------------------------------
    // Grundrechnungsarten
    // ----------------------------------------------------------------------------------- 
    
    public Fraction add(Fraction f)
    {
        if(denominator == f.denominator)
        {
            boolean b = true;
            while(b)
            {
                try
                {
                    numerator = Math.addExact(numerator, f.numerator);
                    b = false;
                }
                catch(ArithmeticException ex)
                {
                    simplify();
                    f.simplify();
                }
            }
        }
        else
        {
            boolean b = true;
            while(b)
            {
                try
                {
                    long l = Math.addExact(Math.multiplyExact(numerator, f.denominator), Math.multiplyExact(f.numerator, denominator));
                    long l2 = Math.multiplyExact(denominator, f.denominator);;
                    numerator = l;
                    denominator = l2;
                    b = false;
                }
                catch(ArithmeticException ex)
                {
                    simplify();
                    f.simplify();
                }
            }
        }
        shortFraction();
        return this;
    }
    
    public Fraction mul(Fraction f)
    {
        boolean b = true;
        while(b)
        {
            try
            {
                long l = Math.multiplyExact(denominator, f.denominator);
                long l2 = Math.multiplyExact(numerator, f.numerator);
                denominator = l;
                numerator = l2;
                b = false;
            }
            catch(ArithmeticException ex)
            {
                simplify();
                f.simplify();
            }
        }
        shortFraction();
        return this;
    }
    
    private Fraction mul(long f)
    {
        boolean b = true;
        while(b)
        {
            try
            {
                numerator = Math.multiplyExact(numerator, f);
                b = false;
            }
            catch(ArithmeticException ex)
            {
                simplify();
            }
        }
        shortFraction();
        return this;
    }
    
    public Fraction sub(Fraction f)
    {
        this.add(f.clone().invertSign());
        return this;
    }
    
    public Fraction div(Fraction f)
    {
        this.mul(f.clone().invert());
        return this;
    }
    
    private Fraction div(long f)
    {
        boolean b = true;
        while(b)
        {
            try
            {
                denominator = Math.multiplyExact(denominator, f);
                b = false;
            }
            catch(ArithmeticException ex)
            {
                simplify();
            }
        }
        shortFraction();
        return this;
    }
    
    // -----------------------------------------------------------------------------------
    // Radizieren, Potenzieren
    // ----------------------------------------------------------------------------------- 
    
    public Fraction squareRoot()
    {
        Fraction f = new Fraction((long) Math.sqrt(this.getDoubleValue()));
        long a = 0;
        long b = 0;
        Fraction f2;
        int k = 0;
        while(true)
        {
            k++;
            f2 = f.clone();
            f.mul(f2).add(this).div(f2.mul(2));
            
            if((k >= 2 && f.denominator == 1) || f.denominator > 281474976710656l || f.numerator == b)
            {
                break;
            }
            
            b = a;
            a = f.numerator;
        }
        tryExactRoot(f);
        numerator = f.numerator;
        denominator = f.denominator;
        if(numerator < 0)
        {
            numerator = -numerator;
        }
        return this;
    }
    
    public Fraction power(Fraction f)
    {
        if(f.denominator == 1)
        {
            power(f.numerator);
            return this;
        }
        throw new ArithmeticException();
    }
    
    private Fraction power(long p)
    {
        if(p < 0)
        {
            p = -p;
            invert();
        }
        else if(p == 1)
        {
            return this;
        }
        long prodn = 1;
        long prodd = 1;
        long n = numerator;
        long d = denominator;
        while(p > 0)
        {
            if((p & 1) == 1)
            {
                prodn *= n;
                prodd *= d;
            }
            p = p >> 1;
            n *= n;
            d *= d;
        }
        numerator = prodn;
        denominator = prodd;
        return this;
    }
    
    private void tryExactRoot(Fraction f)
    {
        Fraction f2 = f.clone();
        if(f2.denominator == 1)
        {
            return;
        }
        int k = 0;
        while(f2.denominator > 1)
        {
            k++;
            f2.simplify();
            if(f2.clone().power(2).equals(this))
            {
                f.numerator = f2.numerator;
                f.denominator = f2.denominator;
                return;
            }
        }
    }
    
    // -----------------------------------------------------------------------------------
    // Invertieren
    // ----------------------------------------------------------------------------------- 
    
    public Fraction invertSign()
    {
        numerator = -numerator;
        return this;
    }
    
    public Fraction invert()
    {
        if(numerator == 0)
        {
            throw new ArithmeticException();
        }
        long helper = numerator;
        if(helper < 0)
        {
            numerator = -denominator;
            denominator = -helper;
        }
        else
        {
            numerator = denominator;
            denominator = helper;
        }
        return this;
    }
    
    // -----------------------------------------------------------------------------------
    // Winkelfunktionen
    // ----------------------------------------------------------------------------------- 
    
    public Fraction sin()
    {
        int factor = 1;
        if(numerator < 0)
        {
            factor = -1;
            numerator = -numerator;
        }
        
        Fraction pi = getPi().mul(2);
        while(this.compareTo(pi) >= 0)
        {
            this.sub(pi);
        }
        
        if(isBetween(new Fraction(0), getPi().div(2)))
        {
            // Nichts
        }
        else if(isBetween(getPi().div(2), getPi()))
        {
            this.invertSign().add(getPi());
        }
        else if(this.compareTo(getPi()) >= 0)
        {
            this.sub(getPi());
            factor = -factor;
        }
        
        
        Fraction quad = this.clone().power(2);
        Fraction value = this.clone();
        Fraction result = this.clone();
        
        int k = 1;
        for(int i = 0; i < 20; i++)
        {
            value.div(i + 2);
            if((i & 1) == 1)
            {
                value.mul(quad);
                k = -k;
                if(k == -1)
                {
                    result.sub(value);
                }
                else
                {
                    result.add(value);
                }
            }
        }
        numerator = factor * result.numerator;
        denominator = result.denominator;
        return this;
    }
    
    public Fraction cos()
    {
        this.add(getPi().div(2));
        return sin();
    }
    
    public Fraction tan()
    {
        sin();
        div(this.clone().cos());
        return this;
    }
    
    // -----------------------------------------------------------------------------------
    // Vereinfachung
    // ----------------------------------------------------------------------------------- 
    
    private void shortFraction()
    {
        if(denominator == 1)
        {
            return;
        }
        long divisor = getGreatestCommonDivisor(numerator, denominator);
        if(divisor != 1)
        {
            denominator /= divisor;
            numerator /= divisor;
        }
    }
    
    private void simplify()
    {
        if(denominator == 1)
        {
            //denominator += 2;
            System.out.println(this);
            throw new ArithmeticException();
        }
        denominator = (denominator & 1) == 1 ? denominator - 1 : denominator;
        numerator = (numerator & 1) == 1 ? numerator - 1 : numerator;
        //System.out.println(denominator + "  WUSI  " + numerator);
        shortFraction();
        //System.out.println(denominator + "  WUSI  " + numerator);
    }
    
    private long getGreatestCommonDivisor(long i, long n)
    {
        if(i == 0)
        {
            return n;
        }
        if(i < 0)
        {
            i = -i;
        }
        if(n < 0)
        {
            n = -n;
        }
        long helper;
        while(true)
        {
            if(i < n)
            {
                helper = i;
                i = n;
                n = helper;
            }
            i = i % n;
            if(i == 0)
            {
                return n;
            }
        }
    }
    
    // -----------------------------------------------------------------------------------
    // Grundobjektcode
    // ----------------------------------------------------------------------------------- 
    
    @Override
    public String toString() 
    {
        return numerator + " / " + denominator;
    }

    @Override
    public Fraction clone()
    {
        return new Fraction(numerator, denominator);
    }

    @Override
    public boolean equals(Object o) 
    {
        if(!(o instanceof Fraction))
        {
            return false;
        }
        Fraction f = (Fraction) o;
        return numerator == f.numerator && denominator == f.denominator;
    }

    @Override
    public int hashCode() 
    {
        int hash = 3;
        hash = 97 * hash + (int) (this.numerator ^ (this.numerator >>> 32));
        hash = 97 * hash + (int) (this.denominator ^ (this.denominator >>> 32));
        return hash;
    }

    @Override
    public int compareTo(Fraction f) 
    {
        if(f.numerator < 0 && numerator >= 0)
        {
            return 1;
        }
        else if(f.numerator >= 0 && numerator < 0)
        {
            return -1;
        }
        else
        {
            long i = f.clone().sub(this).numerator;
            if(i == 0)
            {
                return 0;
            }
            else if(i < 0)
            {
                return 1;
            }
            else
            {
                return -1;
            }
        }
    }
    
    public boolean isBetween(Fraction f1, Fraction f2)
    {
        return this.compareTo(f1) >= 0 && this.compareTo(f2) <= 0;
    }
}
